/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#if defined(__XC)
    #include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>        /* HiTech General Include File */
#endif

#include <stdint.h>        /* For uint8_t definition */

/******************************************************************************/
/* Configuration bits                                                         */
/******************************************************************************/

#pragma config FOSC = INTOSC    // Oscillator Selection->INTOSC oscillator: I/O function on CLKIN pin
#pragma config WDTE = OFF       // Watchdog Timer Enable->WDT disabled
#pragma config PWRTE = OFF      // Power-up Timer Enable->PWRT disabled
#pragma config MCLRE = OFF      // MCLR Pin Function Select->MCLR/VPP pin function is digital input
#pragma config CP = OFF         // Flash Program Memory Code Protection->Program memory code protection is disabled
#pragma config CPD = OFF        // Data Memory Code Protection->Data memory code protection is disabled
#pragma config BOREN = ON       // Brown-out Reset Enable->Brown-out Reset enabled
#pragma config CLKOUTEN = OFF   // Clock Out Enable->CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin
#pragma config IESO = ON        // Internal/External Switchover->Internal/External Switchover mode is enabled
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable->Fail-Safe Clock Monitor is enabled

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection->Write protection off
#pragma config PLLEN = ON       // PLL Enable->4x PLL enabled
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable->Stack Overflow or Underflow will cause a Reset
#pragma config BORV = LO        // Brown-out Reset Voltage Selection->Brown-out Reset Voltage (Vbor), low trip point selected.
#pragma config LVP = OFF        // Low-Voltage Programming Enable->High-voltage on MCLR/VPP must be used for programming

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/

volatile uint16_t   timer_tick;
volatile uint8_t    power_good = 0;
uint16_t            min_count;
uint16_t            max_count;

/******************************************************************************/
/* Main Program (Auto Baud Rate Detection Program for Z80 SIO)                 */
/******************************************************************************/

// 12F1822 Pin Connection
//     +---v---+
// Vdd |1     8| Vss
// SCLK|2     7|~NMI
// CLK |3     6|~RST
// PWR |4     5| RXD
//     +-------+

/* TIMER0 interrupt tick value */
#define IDLE_TIME   1953    // idle tick of time -> 2s
#define TIMEOUT     29297   // timeout of auto detection -> 30s

/* I/O port bit assign */
#define RXD RA2             // RA2 on PIN5 Digital Input(RXD)
#define PWR RA3             // RA3 on PIN4 Digital Input(PWR)
#define RST RA1             // RA1 on PIN6 Digital Output(~RST)
#define NMI RA0             // RA0 on PIN7 Digital Output(~NMI)

/* set PWM parameters */
void set_PWM(uint8_t period, uint8_t tune) {
    OSCTUNE = tune;         // Internal Oscillator Frequency Adjustment
    PR2 = period;           // set PWM Period
    CCPR1L = (uint8_t)(period >> 1); // set PWM width
    TMR2ON = 1;             // Timer2 is on
}

/* get TIMER1 count of mark bit */
void get_mark_bit(uint8_t bits) {
    uint16_t count;
    while(!RXD) if(timer_tick > TIMEOUT || !PWR) return;    // wait space bit
    TMR1ON = 0;                     // stop TIMER1 count
    TMR1H = 0; TMR1L = 0;           // clear TMR1H, TMR1L count
    while(RXD) if(timer_tick > TIMEOUT || !PWR) return;     // wait mark bit
    TMR1ON = 1;                      // start TIMER1 count
    while(!RXD) if(timer_tick > TIMEOUT || !PWR) return;    // wait space bit
    TMR1ON = 0;                     // stop TIMER1 count
    count = ((uint16_t)TMR1H << 8) | TMR1L;
    count >>= bits;                 // normalize count
    if(count < min_count) min_count = count;
    if(count > max_count) max_count = count;
}

/* wait ticks */
void wait_ticks(uint8_t n) {
    volatile uint16_t last;
    for(uint8_t i = 0; i <n; i++) {     // wait n timer_tick
        last = timer_tick;              // save timer_tick
        while(timer_tick == last);      // wait 1 timer_tick
    }
}
        
/* sense POWER and assert RESET */
void sense_power(void) {
    if(!PWR) {
        power_good = 0;     // disable NMI
        wait_ticks(10);     // wait 10mS
        while(!PWR);        // wait power good
        RA1 = 0;            // assert RESET for Z80 CPU & SIO
        wait_ticks(2);      // wait 2mS
        RA1 = 1;            // release RESET
        power_good = 1;     // enable NMI
        wait_ticks(50);     // wait 10mS
        while(!PWR);        // wait power good
    } else RA1 = 1;         // release RESET
}

void interrupt timer_isr(void) {
	if(TMR0IF) {            // timer0 interrupt
        timer_tick++;       // increment time tick in each 1.024 milliseconds
        if(power_good) NMI = ~NMI; else NMI = 1;    // Flip ~NMI
       //if(PWR) { RA1 = 1; } else { RA1 = 0; }
		TMR0IF = 0;         // clear timer0 interrupt
	}
}

/* PIC12F1822 SBC6809 Baud Rate Generator with Auto Speed Detection */

#define PR2_9600        51  // PR2 value for 9600bps
#define PR2_19200       25  // PR2 value for 19200bps
#define PR2_38400       12  // PR2 value for 38400bps
#define PR2_57600       8   // PR2 value for 57600bps
#define PR2_115200      3   // PR2 value for 115200bps

#define MIN_9600        792 // min. TIMER1 count for -5% of 9600bps
#define MIN_19200       396 // min. TIMER1 count for -5% of 19200bps
#define MIN_38400       198 // min. TIMER1 count for -5% of 9600bps
#define MIN_57600       132 // min. TIMER1 count for -5% of 19200bps
#define MIN_115200      66  // min. TIMER1 count for -5% of 9600bps
#define MAX_9600        875 // max. TIMER1 count for +5% of 9600bps
#define MAX_19200       438 // max. TIMER1 count for +5% of 19200bps
#define MAX_38400       219 // max. TIMER1 count for +5% of 38400bps
#define MAX_57600       148 // max. TIMER1 count for +5% of 57600bps
#define MAX_115200      73  // max. TIMER1 count for +5% of 115200bps

/* Auto Speed Detection Sequence
 * no mark bit in 2s and CR in 30s after powerup
 */

void main(void)
{
    /* initialize oscillator */
    OSCCON = 0b11110000;        // PLL Is enabled by CONFIGURATION WORD 1
                                // 8 MHz or 32 MHz HF
                                // Clock determined by CONFIGURATION WORD 1
    
    /* initialize PIN assignment */
    APFCON = 0b00000001;        // PIN2 CCP1 on RA5
    ANSELA = 0;                 // Digital I/O. Pin
    TRISA = 0b11001100;         // RA5 CCP1 on PIN2 Digital Output(SCLK)
                                // RA4 CLKR on PIN3 Digital Output(CLK)
                                // RA3 on PIN4 Digital Input(PWR)
                                // RA2 on PIN5 Digital Input(RXD)
                                // RA1 on PIN6 Digital Output(~RST)
                                // RA0 on PIN7 Digital Output(~NMI)
    WPUA2 = 1;                  // RA2 on PIN5 Pull-up enabled
    
    /* initialize PORT_A */  
    RA1 = 0;                    // ~RST;
    NMI = 1;                    // ~NMI;
    
    /* initialize CLKR */    
    CLKRCON = 0b11010011;       // 1 = Reference clock module is enabled
                                // 1 = Reference clock output is enabled on CLKR pin
                                // 0 = Slew rate limiting is disabled
                                // 10 = Clock outputs duty cycle of 50%
                                // 100 = Base clock value divided by 16 -> 2MHz
                                // 011 = Base clock value divided by 8  -> 4MHz
                                // 010 = Base clock value divided by 4  -> 8MHz

    /* initialize timer */
    OPTION_REG = 0b00000100;    // 0 = Weak pull-ups are enabled by individual WPUx latch values
                                // 0 = Interrupt on falling edge of RB0/INT pin
                                // 0 = Internal instruction cycle clock (FOSC/4)
                                // 0 = Increment on low-to-high transition on T0CKI pin
                                // 0 = Prescaler is assigned to the Timer0 module
                                // 100 = Prescaler Rate Select bits (1 : 32) -> 1.024 milliseconds
    T1GCONbits.TMR1GE = 0;      // always counting
    T1CON = 0b00000000;         // 00 = Timer1 clock source is FOSC/4
                                // 00 = 1:1 Prescale value
                                // 0 = Dedicated Timer1 oscillator circuit disabled
                                // 0 = Synchronize external clock input with system clock (FOSC)
                                // Unimplemented: Read as ?0?
                                // 0 = Stops Timer1
    TMR1H = 0; TMR1L = 0;       // clear TMR1H, TMR1L count
    timer_tick = 0;             // reset timer
    
    /* initialize interrupt setting */ 
    PIR1 = 0;                   // clear peripheral interrupt flag
    PIR2 = 0;                   // clear peripheral interrupt flag
    INTCON = 0b10100000;        // 1 = Enables all active interrupts
                                // 0 = Disables all peripheral interrupts
                                // 1 = Enables the Timer0 interrupt
                                // 0 = Disables the INT external interrupt
                                // 0 = Disables the interrupt-on-change
                                // 0 = TMR0 register did not overflow
                                // 0 = The INT external interrupt did not occur
                                // 0 = None of the interrupt-on-change pins have changed state
    
    /* initialize PWM */
    T2CONbits.T2CKPS = 0b00;    // Prescaler is 1
    CCP1CON = 0b00001100;       // 00 = Single output; P1A modulated; P1B, P1C, P1D assigned as port pins
                                // 00 = These bits are the two LSbs of the PWM duty cycle. The eight MSbs are found in CCPR1L.
                                // 1100 = PWM mode: P1A, P1C active-high; P1B, P1D active-high
    set_PWM(PR2_9600, 0);       // start baud rate generator for 9600bps
    
    /* sensing power */
    sense_power();
    
    /* sensing serial port */
    while(1) {
        timer_tick = 0;
        while(timer_tick < IDLE_TIME) { // wait for 2s
            if(!PWR) break; // if reset is detected
        }
        timer_tick = 0;     // reset timer
        min_count = 65535;  // reset min_count
        max_count = 0;      // reset max_count
        /* CR($0D) 1-0-1011-0000-1
         *           1  1   4bit */
        get_mark_bit(0); // 1 mark bit
        get_mark_bit(0); // 1 mark bit
        get_mark_bit(2); // 4 mark bit
        if(min_count > MIN_9600 && max_count < MAX_9600) {
            set_PWM(PR2_9600, 0); break;
        } else if(min_count > MIN_19200 && max_count < MAX_19200) {
            set_PWM(PR2_19200, 0); break;
        } else if(min_count > MIN_38400 && max_count < MAX_38400) {
            set_PWM(PR2_38400, 0); break;
        } else if(min_count > MIN_57600 && max_count < MAX_57600) {
            set_PWM(PR2_57600, 31); break;
        } else if(min_count > MIN_115200 && max_count < MAX_115200) {
            set_PWM(PR2_115200, 32); break;
        }
        if(timer_tick > TIMEOUT) {   // check timeout
            break;
        }
    }   // done.
    while(1) {  // waiting for reset
        sense_power();
    }
}

